@@ -1,23 +1,9 @@
-/***************************************************************************
- *   Copyright © 2012 Aleix Pol Gonzalez <aleixpol@blue-systems.com>       *
- *   Copyright © 2013 Lukas Appelhans <l.appelhans@gmx.de>                 *
- *                                                                         *
- *   This program is free software; you can redistribute it and/or         *
- *   modify it under the terms of the GNU General Public License as        *
- *   published by the Free Software Foundation; either version 2 of        *
- *   the License or (at your option) version 3 or any later version        *
- *   accepted by the membership of KDE e.V. (or its successor approved     *
- *   by the membership of KDE e.V.), which shall act as a proxy            *
- *   defined in Section 14 of version 3 of the license.                    *
- *                                                                         *
- *   This program is distributed in the hope that it will be useful,       *
- *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
- *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
- *   GNU General Public License for more details.                          *
- *                                                                         *
- *   You should have received a copy of the GNU General Public License     *
- *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
- ***************************************************************************/
+/*
+ *   SPDX-FileCopyrightText: 2012 Aleix Pol Gonzalez <aleixpol@blue-systems.com>
+ *   SPDX-FileCopyrightText: 2013 Lukas Appelhans <l.appelhans@gmx.de>
+ *
+ *   SPDX-License-Identifier: GPL-2.0-only OR GPL-3.0-only OR LicenseRef-KDE-Accepted-GPL
+ */
 
 #include "PackageKitBackend.h"
 #include "PackageKitSourcesBackend.h"
@@ -26,7 +12,7 @@
 #include "AppPackageKitResource.h"
 #include "PKTransaction.h"
 #include "LocalFilePKResource.h"
-#include "TransactionSet.h"
+#include "PKResolveTransaction.h"
 #include <resources/AbstractResource.h>
 #include <resources/StandardBackendUpdater.h>
 #include <resources/SourcesModel.h>
@@ -176,10 +162,9 @@ static DelayedAppStreamLoad loadAppStream(AppStream::Pool* appdata)
 {
     DelayedAppStreamLoad ret;
 
-    QString error;
-    ret.correct = appdata->load(&error);
+    ret.correct = appdata->load();
     if (!ret.correct) {
-        qWarning() << "Could not open the AppStream metadata pool" << error;
+        qWarning() << "Could not open the AppStream metadata pool" << appdata->lastError();
     }
 
     const auto components = appdata->components();
@@ -223,38 +208,12 @@ void PackageKitBackend::reloadPackageList()
                 Q_EMIT passiveMessage(i18n("Please make sure that Appstream is properly set up on your system"));
             });
         }
-        QStringList neededPackages;
-        neededPackages.reserve(data.components.size());
         for (const auto &component: data.components) {
             const auto pkgNames = component.packageNames();
             addComponent(component, pkgNames);
-            neededPackages << pkgNames;
-        }
-        for (auto it = data.missingComponents.constBegin(), itEnd = data.missingComponents.constEnd(); it != itEnd; ++it) {
-            acquireFetching(true);
-            const auto file = it.key();
-            const auto component = it.value();
-            auto trans = PackageKit::Daemon::searchFiles(file);
-            connect(trans, &PackageKit::Transaction::package, this, [trans](PackageKit::Transaction::Info info, const QString &packageID){
-                if (info == PackageKit::Transaction::InfoInstalled)
-                    trans->setProperty("installedPackage", packageID);
-            });
-            connect(trans, &PackageKit::Transaction::finished, this, [this, trans, component](PackageKit::Transaction::Exit status) {
-                const auto pkgidVal = trans->property("installedPackage");
-                if (status == PackageKit::Transaction::ExitSuccess && !pkgidVal.isNull()) {
-                    const auto pkgid = pkgidVal.toString();
-                    auto res = addComponent(component, {PackageKit::Daemon::packageName(pkgid)});
-                    res->clearPackageIds();
-                    res->addPackageId(PackageKit::Transaction::InfoInstalled, pkgid, true);
-                }
-                acquireFetching(false);
-            });
         }
 
-        if (!neededPackages.isEmpty()) {
-            neededPackages.removeDuplicates();
-            resolvePackages(neededPackages);
-        } else {
+        if (data.components.isEmpty()) {
             qCDebug(LIBDISCOVER_BACKEND_LOG) << "empty appstream db";
             if (PackageKit::Daemon::backendName() == QLatin1String("aptcc") || PackageKit::Daemon::backendName().isEmpty()) {
                 checkForUpdates();
@@ -274,11 +233,11 @@ AppPackageKitResource* PackageKitBackend::addComponent(const AppStream::Componen
     Q_ASSERT(isFetching());
     Q_ASSERT(!pkgNames.isEmpty());
 
-
-    AppPackageKitResource* res = qobject_cast<AppPackageKitResource*>(m_packages.packages[component.id()]);
+    auto& resPos = m_packages.packages[component.id()];
+    AppPackageKitResource* res = qobject_cast<AppPackageKitResource*>(resPos);
     if (!res) {
         res = new AppPackageKitResource(component, pkgNames.at(0), this);
-        m_packages.packages[component.id()] = res;
+        resPos = res;
     } else {
         res->clearPackageIds();
     }
@@ -294,16 +253,15 @@ AppPackageKitResource* PackageKitBackend::addComponent(const AppStream::Componen
 
 void PackageKitBackend::resolvePackages(const QStringList &packageNames)
 {
-    PackageKit::Transaction * tArch = PackageKit::Daemon::resolve(packageNames, PackageKit::Transaction::FilterArch);
-    connect(tArch, &PackageKit::Transaction::package, this, &PackageKitBackend::addPackageArch);
-    connect(tArch, &PackageKit::Transaction::errorCode, this, &PackageKitBackend::transactionError);
+    if (!m_resolveTransaction) {
+        m_resolveTransaction = new PKResolveTransaction(this);
+        connect(m_resolveTransaction, &PKResolveTransaction::allFinished, this, &PackageKitBackend::getPackagesFinished);
+        connect(m_resolveTransaction, &PKResolveTransaction::started, this, [this] {
+            m_resolveTransaction = nullptr;
+        });
+    }
 
-    PackageKit::Transaction * tNotArch = PackageKit::Daemon::resolve(packageNames, PackageKit::Transaction::FilterNotArch);
-    connect(tNotArch, &PackageKit::Transaction::package, this, &PackageKitBackend::addPackageNotArch);
-    connect(tNotArch, &PackageKit::Transaction::errorCode, this, &PackageKitBackend::transactionError);
-
-    TransactionSet* merge = new TransactionSet({tArch, tNotArch});
-    connect(merge, &TransactionSet::allFinished, this, &PackageKitBackend::getPackagesFinished);
+    m_resolveTransaction->addPackageNames(packageNames);
 }
 
 void PackageKitBackend::fetchUpdates()
@@ -320,8 +278,8 @@ void PackageKitBackend::fetchUpdates()
     m_hasSecurityUpdates = false;
 
     m_updater->setProgressing(true);
-    
-    fetchingUpdatesProgressChanged();
+
+    Q_EMIT fetchingUpdatesProgressChanged();
 }
 
 void PackageKitBackend::addPackageArch(PackageKit::Transaction::Info info, const QString& packageId, const QString& summary)
@@ -355,13 +313,6 @@ void PackageKitBackend::addPackage(PackageKit::Transaction::Info info, const QSt
 
 void PackageKitBackend::getPackagesFinished()
 {
-    for(auto it = m_packages.packages.cbegin(); it != m_packages.packages.cend(); ++it) {
-        auto pkr = qobject_cast<PackageKitResource*>(it.value());
-        if (pkr->packages().isEmpty()) {
-//             qWarning() << "Failed to find package for" << it.key();
-            m_packagesToDelete += pkr;
-        }
-    }
     includePackagesToAdd();
 }
 
@@ -461,23 +412,101 @@ QList<AppStream::Component> PackageKitBackend::componentsById(const QString& id)
     return m_appdata->componentsById(id);
 }
 
+static const auto needsResolveFilter = [] (AbstractResource* res) { return res->state() == AbstractResource::Broken; };
+static const auto installedFilter = [] (AbstractResource* res) { return res->state() >= AbstractResource::Installed; };
+
+class PKResultsStream : public ResultsStream
+{
+public:
+    PKResultsStream(PackageKitBackend* backend, const QString &name)
+        : ResultsStream(name)
+        , backend(backend)
+    {}
+
+    PKResultsStream(PackageKitBackend* backend, const QString &name, const QVector<AbstractResource*> &resources)
+        : ResultsStream(name)
+        , backend(backend)
+    {
+        QTimer::singleShot(0, this, [resources, this] () {
+            if (!resources.isEmpty())
+                setResources(resources);
+            finish();
+        });
+    }
+
+    void setResources(const QVector<AbstractResource*> &res)
+    {
+        const auto toResolve = kFilter<QVector<AbstractResource*>>(res, needsResolveFilter);
+        if (!toResolve.isEmpty())
+            backend->resolvePackages(kTransform<QStringList>(toResolve, [] (AbstractResource* res) { return res->packageName(); }));
+        Q_EMIT resourcesFound(res);
+    }
+private:
+    PackageKitBackend* const backend;
+};
+
 ResultsStream* PackageKitBackend::search(const AbstractResourcesBackend::Filters& filter)
 {
     if (!filter.resourceUrl.isEmpty()) {
         return findResourceByPackageName(filter.resourceUrl);
     } else if (!filter.extends.isEmpty()) {
-        const auto ext = kTransform<QVector<AbstractResource*>>(m_packages.extendedBy.value(filter.extends), [](AppPackageKitResource* a){ return a; });
-        return new ResultsStream(QStringLiteral("PackageKitStream-extends"), ext);
+        auto stream = new PKResultsStream(this, QStringLiteral("PackageKitStream-extends"));
+        auto f = [this, filter, stream] {
+            const auto resources = kTransform<QVector<AbstractResource*>>(m_packages.extendedBy.value(filter.extends), [](AppPackageKitResource* a){ return a; });
+            if (!resources.isEmpty()) {
+                stream->setResources(resources);
+            }
+        };
+        runWhenInitialized(f, stream);
+        return stream;
+    } else if (filter.state == AbstractResource::Upgradeable) {
+        return new ResultsStream(QStringLiteral("PackageKitStream-upgradeable"), kTransform<QVector<AbstractResource*>>(upgradeablePackages())); //No need for it to be a PKResultsStream
+    } else if (filter.state == AbstractResource::Installed) {
+        auto stream = new PKResultsStream(this, QStringLiteral("PackageKitStream-installed"));
+        auto f = [this, stream] {
+            const auto toResolve = kFilter<QVector<AbstractResource*>>(m_packages.packages, needsResolveFilter);
+
+            if (!toResolve.isEmpty()) {
+                resolvePackages(kTransform<QStringList>(toResolve, [] (AbstractResource* res) { return res->packageName(); }));
+                connect(m_resolveTransaction, &PKResolveTransaction::allFinished, this, [stream, toResolve] {
+                    const auto resolved = kFilter<QVector<AbstractResource*>>(toResolve, installedFilter);
+                    if (!resolved.isEmpty())
+                        Q_EMIT stream->resourcesFound(resolved);
+                    stream->finish();
+                });
+            }
+
+            const auto resolved = kFilter<QVector<AbstractResource*>>(m_packages.packages, installedFilter);
+            if (!resolved.isEmpty()) {
+                QTimer::singleShot(0, this, [resolved, toResolve, stream] () {
+                    if (!resolved.isEmpty())
+                        Q_EMIT stream->resourcesFound(resolved);
+
+                    if (toResolve.isEmpty())
+                        stream->finish();
+            });
+            }
+        };
+        runWhenInitialized(f, stream);
+        return stream;
     } else if (filter.search.isEmpty()) {
-        return new ResultsStream(QStringLiteral("PackageKitStream-all"), kFilter<QVector<AbstractResource*>>(m_packages.packages, [](AbstractResource* res) { return res->type() != AbstractResource::Technical && !qobject_cast<PackageKitResource*>(res)->extendsItself(); }));
+        auto stream = new PKResultsStream(this, QStringLiteral("PackageKitStream-all"));
+        auto f = [this, filter, stream] {
+            auto resources = kFilter<QVector<AbstractResource*>>(m_packages.packages, [](AbstractResource* res) { return res->type() != AbstractResource::Technical && !qobject_cast<PackageKitResource*>(res)->extendsItself(); });
+            if (!resources.isEmpty()) {
+                stream->setResources(resources);
+            }
+        };
+        runWhenInitialized(f, stream);
+        return stream;
     } else {
-        auto stream = new ResultsStream(QStringLiteral("PackageKitStream-search"));
+        auto stream = new PKResultsStream(this, QStringLiteral("PackageKitStream-search"));
         const auto f = [this, stream, filter] () {
             const QList<AppStream::Component> components = m_appdata->search(filter.search);
             const QStringList ids = kTransform<QStringList>(components, [](const AppStream::Component& comp) { return comp.id(); });
             if (!ids.isEmpty()) {
                 const auto resources = kFilter<QVector<AbstractResource*>>(resourcesByPackageNames<QVector<AbstractResource*>>(ids), [](AbstractResource* res){ return !qobject_cast<PackageKitResource*>(res)->extendsItself(); });
-                Q_EMIT stream->resourcesFound(resources);
+                stream->setResources(resources);
             }
 
             PackageKit::Transaction * tArch = PackageKit::Daemon::resolve(filter.search, PackageKit::Transaction::FilterArch);
@@ -491,22 +520,27 @@ ResultsStream* PackageKitBackend::search(const AbstractResourcesBackend::Filters
                     const auto packageId = stream->property("packageId");
                     if (!packageId.isNull()) {
                         const auto res = resourcesByPackageNames<QVector<AbstractResource*>>({PackageKit::Daemon::packageName(packageId.toString())});
-                        Q_EMIT stream->resourcesFound(kFilter<QVector<AbstractResource*>>(res, [ids](AbstractResource* res){ return !ids.contains(res->appstreamId()); }));
+                        stream->setResources(kFilter<QVector<AbstractResource*>>(res, [ids](AbstractResource* res){ return !ids.contains(res->appstreamId()); }));
                     }
                 }
                 stream->finish();
             }, Qt::QueuedConnection);
         };
-        if (!m_appstreamInitialized) {
-            connect(this, &PackageKitBackend::loadedAppStream, stream, f);
-        } else {
-            QTimer::singleShot(0, this, f);
-        }
+        runWhenInitialized(f, stream);
         return stream;
     }
 }
 
-ResultsStream * PackageKitBackend::findResourceByPackageName(const QUrl& url)
+void PackageKitBackend::runWhenInitialized(const std::function<void ()>& f, QObject* stream)
+{
+    if (!m_appstreamInitialized) {
+        connect(this, &PackageKitBackend::loadedAppStream, stream, f);
+    } else {
+        QTimer::singleShot(0, this, f);
+    }
+}
+
+PKResultsStream * PackageKitBackend::findResourceByPackageName(const QUrl& url)
 {
     if (url.isLocalFile()) {
         QMimeDatabase db;
@@ -517,7 +551,7 @@ ResultsStream * PackageKitBackend::findResourceByPackageName(const QUrl& url)
             || mime.inherits(QStringLiteral("application/x-zstd-compressed-tar"))
             || mime.inherits(QStringLiteral("application/x-xz-compressed-tar"))
         ) {
-            return new ResultsStream(QStringLiteral("PackageKitStream-localpkg"), QVector<AbstractResource*>{ new LocalFilePKResource(url, this)});
+            return new PKResultsStream(this, QStringLiteral("PackageKitStream-localpkg"), { new LocalFilePKResource(url, this)});
         }
     } else if (url.scheme() == QLatin1String("appstream")) {
         static const QMap<QString, QString> deprecatedAppstreamIds = {
@@ -534,7 +568,7 @@ ResultsStream * PackageKitBackend::findResourceByPackageName(const QUrl& url)
         if (appstreamIds.isEmpty())
             Q_EMIT passiveMessage(i18n("Malformed appstream url '%1'", url.toDisplayString()));
         else {
-            auto stream = new ResultsStream(QStringLiteral("PackageKitStream-appstream-url"));
+            auto stream = new PKResultsStream(this, QStringLiteral("PackageKitStream-appstream-url"));
             const auto f = [this, appstreamIds, stream] () {
                 AbstractResource* pkg = nullptr;
 
@@ -548,8 +582,10 @@ ResultsStream * PackageKitBackend::findResourceByPackageName(const QUrl& url)
 
                 for (auto it = m_packages.packages.constBegin(), itEnd = m_packages.packages.constEnd(); it != itEnd; ++it) {
                     const bool matches = kContains(allAppStreamIds, [&it] (const QString& id) {
+                        static const QLatin1String desktopPostfix(".desktop");
                         return it.key().compare(id, Qt::CaseInsensitive) == 0 ||
-                              (id.endsWith(QLatin1String(".desktop")) && id.compare(it.key()+QLatin1String(".desktop"), Qt::CaseInsensitive) == 0);
+                              //doing (id == id.key()+".desktop") without allocating
+                              (id.size() == (desktopPostfix.size() + it.key().size()) && id.endsWith(desktopPostfix) && id.startsWith(it.key(), Qt::CaseInsensitive));
                     });
                     if (matches) {
                         pkg = it.value();
@@ -557,20 +593,16 @@ ResultsStream * PackageKitBackend::findResourceByPackageName(const QUrl& url)
                     }
                 }
                 if (pkg)
-                    Q_EMIT stream->resourcesFound({pkg});
+                    stream->setResources({pkg});
                 stream->finish();
     //             if (!pkg)
     //                 qCDebug(LIBDISCOVER_BACKEND_LOG) << "could not find" << host << deprecatedHost;
             };
-            if (!m_appstreamInitialized) {
-                connect(this, &PackageKitBackend::loadedAppStream, stream, f);
-            } else {
-                QTimer::singleShot(0, this, f);
-            }
+            runWhenInitialized(f, stream);
             return stream;
         }
     }
-    return new ResultsStream(QStringLiteral("PackageKitStream-unknown-url"), {});
+    return new PKResultsStream(this, QStringLiteral("PackageKitStream-unknown-url"), {});
 }
 
 bool PackageKitBackend::hasSecurityUpdates() const
